import math

import numpy as np
import pandas as pd
from rich.progress import Progress, TextColumn, BarColumn, TaskProgressColumn, MofNCompleteColumn
from rich import print

MSSQL_INT_TYPES = ['bigint', 'int', 'smallint', 'tinyint']
MSSQL_FLOAT_TYPES = ['decimal', 'numeric', 'float']
MSSQL_STR_TYPES = ['varchar', 'nvarchar', 'char', 'nchar']
MSSQL_DATE_TYPES = ['date', 'datetime', 'datetime2']
NUMPY_INT_TYPES = [np.int_, np.int64, np.int32, np.int8, 'Int64']
NUMPY_FLOAT_TYPES = [np.float64, np.float32, np.float16, 'Float64']
NUMPY_STR_TYPES = [np.str_, np.object_, 'string']
NUMPY_BOOL_TYPES = [np.bool_, np.True_, np.False_, pd.BooleanDtype, 'boolean']


def insert_to_mssql_db(column_string, cursor, data_list, location, values):
    value_list = " union ".join(['select {}'.format(value) for value in values])
    execute_query = (
        f"insert into {location} ({column_string}) {value_list}"
    )
    try:
        cursor.execute(execute_query, data_list)
    except Exception as e:
        print(execute_query)
        print(data_list)
        raise e


class Loader:
    @staticmethod
    def insert_to_mssql_table(cursor, df: pd.DataFrame, schema: str, table: str):
        column_list = df.columns.tolist()
        column_list = [f'[{column}]' for column in column_list]
        column_string = ", ".join(column_list)
        location = f"{schema}.[{table}]"

        row_values = []
        for column in df.columns:
            series = df[column]
            series_type = series.dtype
            str_column = series.apply(str)
            max_size = str_column.str.len().max()
            if max_size > 256:
                row_values.append('cast ( ? as nvarchar(max))')
            else:
                row_values.append('?')
            # switches from numpy class to python class for bool float and int
            if series_type in NUMPY_BOOL_TYPES or series_type in NUMPY_INT_TYPES or series_type in NUMPY_FLOAT_TYPES:
                df[column] = series.tolist()
        row_value_list = ", ".join(row_values)
        df = df.replace({np.nan: None})
        with Progress(TextColumn("[progress.description]{task.description}"), BarColumn(), TaskProgressColumn(),
                      MofNCompleteColumn()) as progress:
            total = df.shape[0]
            values = []
            data_list = []
            data_count = 0
            row_count = 0
            upload_task = progress.add_task(f'loading {table}', total=total)
            for row in df.itertuples(index=False, name=None):
                row_size = len(row)
                row_count += 1
                data_count += row_size
                values.append(row_value_list)

                data_list.extend(row)
                next_size = data_count + row_size
                if next_size >= 2000:
                    insert_to_mssql_db(column_string, cursor, data_list, location, values)
                    progress.update(upload_task, advance=row_count)
                    values = []
                    data_list = []
                    data_count = 0
                    row_count = 0
            if row_count > 0:
                insert_to_mssql_db(column_string, cursor, data_list, location, values)
                progress.update(upload_task, advance=row_count)

    @staticmethod
    def validate_mssql_upload(connection, df: pd.DataFrame, schema: str, table: str):
        get_column_info_query = (
            f'select COLUMN_NAME, DATA_TYPE, CHARACTER_MAXIMUM_LENGTH, NUMERIC_PRECISION '
            f'from INFORMATION_SCHEMA.columns '
            f'where table_schema = \'{schema}\' and table_name = \'{table}\'')
        column_info_df = pd.read_sql(get_column_info_query, connection)
        # make sure df doesn't have any extra columns
        df_columns = df.columns.tolist()
        db_columns = column_info_df['COLUMN_NAME'].tolist()
        new_columns = np.setdiff1d(df_columns, db_columns)
        if len(new_columns) > 0:
            extra_columns_string = ", ".join(new_columns)
            type_mismatch_error_message = \
                f'The table {schema}.{table} is missing the following columns: {extra_columns_string} '
            raise ExtraColumnsException(type_mismatch_error_message)
        # make sure column types match up
        type_mismatch_columns = []
        truncated_columns = []
        for column in df_columns:
            db_column_info = column_info_df[column_info_df['COLUMN_NAME'] == column]
            db_column_data_type = db_column_info.iloc[0]['DATA_TYPE']
            df_column_data_type = df[column].dtype
            db_column_numeric_precision = db_column_info.iloc[0]['NUMERIC_PRECISION']
            db_column_string_length = db_column_info.iloc[0]['CHARACTER_MAXIMUM_LENGTH']
            type_mismatch_error_message = (f'{column} in dataframe is of type {df_column_data_type} '
                                           f'while the database expects a type of {db_column_data_type}')
            if df_column_data_type in NUMPY_INT_TYPES:
                if db_column_data_type not in MSSQL_INT_TYPES:
                    type_mismatch_columns.append(type_mismatch_error_message)
                    continue
                df_numeric_precision = int(math.log10(df[column].max())) + 1
                if df_numeric_precision > db_column_numeric_precision:
                    truncate_error_message = (f'{column} needs a minimum of {df_numeric_precision} '
                                              f'precision to be inserted')
                    truncated_columns.append(truncate_error_message)
                    continue

            elif df_column_data_type in NUMPY_FLOAT_TYPES:
                if db_column_data_type not in MSSQL_FLOAT_TYPES:
                    type_mismatch_columns.append(type_mismatch_error_message)
                    continue
                df_numeric_precision = int(math.log10(df[column].max())) + 1
                if df_numeric_precision > db_column_numeric_precision:
                    truncate_error_message = (f'{column} needs a minimum of {df_numeric_precision} '
                                              f'precision to be inserted')
                    truncated_columns.append(truncate_error_message)
                    continue

            elif df_column_data_type is np.datetime64:
                if db_column_data_type not in MSSQL_DATE_TYPES:
                    type_mismatch_columns.append(type_mismatch_error_message)
                    continue
            elif df_column_data_type in NUMPY_STR_TYPES:
                if db_column_data_type not in MSSQL_STR_TYPES:
                    type_mismatch_columns.append(type_mismatch_error_message)
                    continue
                df_max_string_length = df[column].str.len().max()
                if df_max_string_length > db_column_string_length:
                    truncate_error_message = (f'{column} needs a minimum of {df_max_string_length} '
                                              f'size to be inserted')
                    truncated_columns.append(truncate_error_message)
                    continue
        if len(truncated_columns) > 0 or len(type_mismatch_columns) > 0:
            error_message = '\n'.join(type_mismatch_columns) + '\n'.join(truncated_columns)
            raise ColumnDataException(error_message)


class ExtraColumnsException(Exception):
    pass


class ColumnDataException(Exception):
    pass



import hashlib
import re
import numpy as np
import pandas as pd
from dateutil import parser
from rich import print


def clean_bool(dirty_bool):
    if dirty_bool is None:
        return
    dirty_bool = str(dirty_bool).lower()
    if dirty_bool in ('y', 'yes', 't', 'true', 'on', '1'):
        return True
    elif dirty_bool in ('n', 'no', 'f', 'false', 'off', '0'):
        return False
    else:
        raise ValueError("invalid truth value %r" % (dirty_bool,))


def clean_float(dirty_float):
    if dirty_float is None:
        return
    dirty_float = str(dirty_float)
    cleaned = dirty_float.replace(',', '').replace('$', '').replace('%', '')
    return float(cleaned)


def clean_date(dirty_date):
    if dirty_date is None or dirty_date is np.nan:
        return
    dirty_date = str(dirty_date).strip()
    return parser.parse(dirty_date)


def clean_int(dirty_int):
    if dirty_int is None or np.isnan(dirty_int):
        return
    if dirty_int == int(dirty_int):
        return int(dirty_int)
    raise ValueError


def hash_str(not_hashed):
    hashed = hashlib.sha1(str(not_hashed).encode())
    return hashed.hexdigest()


class Cleaner:

    @staticmethod
    def column_names_to_snake_case(df: pd.DataFrame):
        column_names = df.columns.to_list()
        clean_names = []
        for name in column_names:
            # Remove / Replace non-alpha-numeric characters
            name = (str(name)
                    .strip()
                    .replace('?', '')
                    .replace('(', '')
                    .replace(')', '')
                    .replace('\\', '')
                    .replace(',', '')
                    .replace('#', 'Num')
                    .replace('$', 'Dollars'))
            # add in underscore before a capital letter or number
            name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
            name = re.sub('([a-z0-9])([A-Z])', r'\1_\2', name).lower()
            # replace other word-splitters with underscores
            name = (name
                    .replace('.', '_')
                    .replace(':', '_')
                    .replace(' ', '_')
                    .replace('-', '_')
                    .replace('___', '_')
                    .replace('__', '_')
                    .strip('_'))
            clean_names.append(name)
        df.columns = clean_names

    @staticmethod
    def clean_numbers(df: pd.DataFrame):
        for column, values in df.items():
            try:
                clean_floats = values.apply(clean_float)
                df[column] = clean_floats
                print(f'{column} has been cast to float')
                df[column] = clean_floats.apply(clean_int)
                print(f'{column} has been cast to int')
            except (ValueError, TypeError):
                continue
        return df

    @staticmethod
    def clean_dates(df: pd.DataFrame):
        for column, values in df.items():
            try:
                df[column] = values.apply(clean_date)
                print(f'{column} has been cast to datetime')
            except (parser.ParserError, OverflowError):
                continue
        return df

    @staticmethod
    def clean_bools(df: pd.DataFrame):
        for column, values in df.items():
            try:
                df[column] = values.apply(clean_bool)
                print(f'{column} has been cast to bool')
            except ValueError:
                continue
        return df

    @staticmethod
    def clean_all(df: pd.DataFrame):
        for column, values in df.items():
            try:
                clean_floats = values.apply(clean_float)
                df[column] = clean_floats
                print(f'{column} has been cast to float')
                df[column] = clean_floats.apply(clean_int)
                print(f'{column} has been cast to int')
            except (ValueError, TypeError):
                pass
            try:
                df[column] = values.apply(clean_date)
                print(f'{column} has been cast to datetime')
            except (parser.ParserError, OverflowError):
                pass
            try:
                df[column] = values.apply(clean_bool)
                print(f'{column} has been cast to bool')
            except ValueError:
                pass
        df = df.convert_dtypes()
        return df

    @staticmethod
    def generate_hash_column(df: pd.DataFrame, columns_to_hash: list[str], new_column_name: str):
        df[new_column_name] = ""
        for column in columns_to_hash:
            df[new_column_name] += df[column].apply(str)
        df[new_column_name] = df[new_column_name].apply(hash_str)
        return df

    @staticmethod
    def coalesce_columns(df: pd.DataFrame, columns_to_coalesce: list[str], new_column_name: str, drop: bool = False):
        df[new_column_name] = df[columns_to_coalesce[1]]
        for column in columns_to_coalesce:
            df[new_column_name] = df[new_column_name].combine_first(df[column])
        if drop:
            df = df.drop(columns=columns_to_coalesce)
        return df
